/*
 * TRAN.C - Translate COBOL File <--> DM File.
 *
 * Usage:
 *
 *    tran [-r] [-p] <tranFile> <cobolFile> <dmFile>
*/

#include "env.h"  /* must appear before first C statement */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hash.h"
#include "dmio.h"
#include "dynabuf.h"
#include "tran.h"

#define even(i)   (0 == ((i) % 2))
#define _MAX_FNAME  256

typedef struct
{
	char *name;
#if defined(IBM_370_PARM_NAME_BUG)
	bool (*toDm)(FILE *, byte *, tranQEntry *);
	bool (*toCobol)(byte *, char *, tranQEntry *);
#else
	bool (*toDm)(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
	bool (*toCobol)(byte *pbRecBuf, char *strFld, tranQEntry *ptq);
#endif
} convFuncType;

typedef struct
{
	FILE *pF;
	char strFileName[MAX_FILENAME_LEN];
} isamData;

static QUEUE qIsamFiles;
#if defined(IBM)
static bool  pc_mode = false;
#else
static bool  pc_mode = true;
#endif

static bool copyToDm(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
static bool stringToIsam(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
static bool writeIsamRecord( isamData *pID, byte *pbRecBuf,
                             attribute *pa, byte *val, size_t valLen );
static bool packToIsam(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
static bool zoneToFloat(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
static bool zoneToInt(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
static bool packToFloat(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
static bool packToInt(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);
static bool unsupported(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq);

static bool copyToCobol(byte *pbRecBuf, char *strFld, tranQEntry *ptq);
static bool floatToZone(byte *pbRecBuf, char *strFld, tranQEntry *ptq);
static bool intToZone(byte *pbRecBuf, char *strFld, tranQEntry *ptq);
static bool floatToPack(byte *pbRecBuf, char *strFld, tranQEntry *ptq);
static bool intToPack(byte *pbRecBuf, char *strFld, tranQEntry *ptq);
static bool unsupportedToCobol(byte *pbRecBuf, char *strFld, tranQEntry *ptq);

static convFuncType convFuncTable[] =
{
	{	"copy", 			copyToDm,		copyToCobol				},
	{	"alphaToStr",	stringToIsam,	copyToCobol				},
	{	"zoneToFloat",	zoneToFloat,	floatToZone				},
	{	"zoneToInt",	zoneToInt,		intToZone				},
	{	"packToFloat",	packToFloat,	floatToPack				},
	{	"packToInt",	packToInt,		intToPack				},
	{	"packToStr",	packToIsam,		intToPack				},
	{  "unsupported", unsupported,   unsupportedToCobol	},
	{	""																	}
};

static int getFld(FILE *fpDmFile, char *strFld, int bufSize, int *piLine,
	int *piCurrFld);
static void readDmHeader(FILE *f, QUEUE q);
static void setConvFuncPtrs(tranQEntry *ptq, char strConvFuncName[]);
static void setFldCnt(QUEUE q, char DmName[], int iFldCt);
static QUEUE sortQueueIntoFieldOrder(QUEUE q);
static void usage(void);
static void writeDmHeader(FILE *f, QUEUE q);
static bool startConversions(void);
static bool endConversions(void);
static void fieldAdjust(char **ppc, size_t *ps);
static bool trnFileRead(FILE *fpTrn, char *strFname, QUEUE *pq, byte **ppbRecBuf);
static bool chkFldsAvail(QUEUE q);
static bool chkFldOverlap(QUEUE q, byte *pbRecBuf);
static void ffwrite(byte *pbBuf, size_t sLen, FILE *pf);
static tranQEntry *getTranQEntry(QUEUE q, char *strFldName);
static int ksIntCompn(ulong ul0, char *s, size_t n);
static int packToInteger(byte *pbRecBuf, tranQEntry *ptq);
static int zoneToInteger(byte *pbRecBuf, tranQEntry *ptq);
static int compTQEbyFldOrder(tranQEntry *ptq0, tranQEntry *ptq1);
static FILE *openKeyedStringFile( char *fileName );
static char *ebc2asc( char *pc, size_t iLen );

static char *ProgName;
static DYNA_BUFFER db;
static byte *PcIsamRec;
static byte *prevIsamRecKey;
static char space = ' ';
static struct
{
	ulong ulRecLen;
	ulong ulMaxIsamRecLen;
	key k;
	size_t keyPrefixLen;
	ulong ulFldLen;
} TrnHdr;
static QUEUE q;	/* QUEUE of trnQEntry */

unsigned char ebc2asc_tbl[256] = {

   '\x00','\x01','\x02','\x03','\x00','\x09','\x20','\x7F',
   '\x20','\x20','\x20','\x0B','\x0C','\x0D','\x0E','\x0F',
   '\x10','\x11','\x12','\x13','\x20','\x20','\x08','\x20',
   '\x18','\x19','\x20','\x20','\x1C','\x1D','\x1E','\x1F',
   '\x20','\x20','\x1C','\x20','\x20','\x0A','\x17','\x1B',
   '\x20','\x20','\x20','\x20','\x20','\x05','\x06','\x07',
   '\x00','\x00','\x16','\x20','\x20','\x20','\x20','\x04',
   '\x20','\x20','\x20','\x20','\x14','\x15','\x00','\x1A',
   '\x5F','\x20','\x20','\x20','\x20','\x20','\x20','\x20', 
   '\x20','\x20','\x9B','\x2E','\x3C','\x28','\x2B','\x7C',
   '\x26','\x20','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x20','\x20','\x21','\x24','\x2A','\x29','\x3B','\xBF',
   '\x2D','\x2F','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x20','\x20','\x7C','\x2C','\x25','\x5F','\x3E','\x3F',
   '\x20','\x20','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x20','\x60','\x3A','\x23','\x40','\x27','\x3D','\x22',
   '\x20','\x61','\x62','\x63','\x64','\x65','\x66','\x67',
   '\x68','\x69','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x20','\x6A','\x6B','\x6C','\x6D','\x6E','\x6F','\x70',
   '\x71','\x72','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x20','\x7E','\x73','\x74','\x75','\x76','\x77','\x78',
   '\x79','\x7A','\x20','\x20','\x20','\x5B','\x20','\x20',
   '\x20','\x20','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x20','\x20','\x20','\x20','\x20','\x5D','\x20','\x20',
   '\x7B','\x41','\x42','\x43','\x44','\x45','\x46','\x47',
   '\x48','\x49','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x7D','\x4A','\x4B','\x4C','\x4D','\x4E','\x4F','\x50',
   '\x51','\x52','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x5C','\x20','\x53','\x54','\x55','\x56','\x57','\x58',
   '\x59','\x5A','\x20','\x20','\x20','\x20','\x20','\x20',
   '\x30','\x31','\x32','\x33','\x34','\x35','\x36','\x37',
   '\x38','\x39','\x7C','\x20','\x20','\x20','\x20','\x20'
};

int main(int argc, char *argv[])
{
	FILE *fpTrn, *fpCobolFile, *fpDmFile;
	byte *pbRecBuf;
	int iLine, iCurrFld, counter;
	size_t size;
	tranQEntry *ptq;
	bool fReverse = false, fLineHasError = false;
        bool fFldTransErr = false;
	char *pcFld;
/*******/
        int bl, rl;
/*******/

	ProgName = argv[0];

	/* parse command line options */
   while ( --argc > 0  &&  (*++argv)[0] == '-' )
      switch((*argv)[1])
		{
		case 'r':
			fReverse = true;
			break;

	       case 'p':
			pc_mode = true;
			break;

               case '?':
	        
	        default:
			usage();
      }

	if(3 != argc)
		usage();

	if(NULL == (fpTrn = fopen(argv[0], "r")))
	{
		fprintf(stderr,"%s: cannot open '%s'\n", ProgName, argv[0]);
		return(1);
	}

	if(NULL == (fpCobolFile = fopen(argv[1], fReverse ? "wb" : "rb")))
	{
		fprintf(stderr,"%s: cannot open '%s'\n", ProgName, argv[1]);
		return(1);
	}

	if(NULL == (fpDmFile = fopen(argv[2], fReverse ? "r" : "w")))
	{
		fprintf(stderr,"%s: cannot open '%s'\n", ProgName, argv[2]);
		return(1);
	}

	if(!trnFileRead(fpTrn, argv[0], &q, &pbRecBuf))
	{
		fprintf(stderr, "%s: cannot read translation request file header\n",
			ProgName);
		exit(1);
	}

   /* +1 for terminal newline character */
	size = TrnHdr.keyPrefixLen + (size_t) TrnHdr.ulMaxIsamRecLen + 1;
	PcIsamRec = malloc(size);
   if(NULL == PcIsamRec)
	{
		fprintf(stderr, "unable to allocate buffer of size %u\n", size);
		exit(1);
	}
	prevIsamRecKey = malloc( (size_t) TrnHdr.k.ulLen );
   if( prevIsamRecKey == NULL )
	{
		fprintf(stderr, "unable to allocate Key buffer of size %lu\n",
                       TrnHdr.k.ulLen );
		exit(1);
	}
   memset( prevIsamRecKey, '\0', (size_t) TrnHdr.k.ulLen );

	pcFld = malloc( (size_t) TrnHdr.ulFldLen );
   if ( pcFld == NULL )
	{
		fprintf(stderr, "unable to allocate field buffer of size %u\n",
			(size_t) TrnHdr.ulFldLen);
		exit(1);
	}

   (void) startConversions();

   if(fReverse)
	{
		QUEUE qSorted;
		int iPrevLine, iFldsTranslated;
		QUEUE_ITERATOR qitTq;

		/* read DM file header to get field offsets */
		readDmHeader(fpDmFile, q);
		if(!chkFldsAvail(q))
		{
			exit(1);
		}
		/* sort translations into field offset order */
		qSorted = sortQueueIntoFieldOrder(q);
		qdestroy(&q);
		q = qSorted;
#if !defined(NDEBUG)
	FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
		printf("%s  fldOff %d  %ld  %ld\n", ptq->DmName, ptq->iFldOffset,
			ptq->ste.length, ptq->ste.offset);
#endif

		iPrevLine = iLine = 1;
		iCurrFld = 0;	/* currFld numbered from 1 to N, like field offset */
		iFldsTranslated = 0;
		fLineHasError = false;
		qitTq = qitcreat(q);
		ptq = qitnext(qitTq);
		if(NULL == ptq)
		{
			fprintf(stderr, "%s: (internal) Queue header lost\n", ProgName);
			exit(1);
		}
		while(EOF != getFld(fpDmFile, pcFld, (size_t) TrnHdr.ulFldLen,
			&iLine, &iCurrFld))
		{
#if !defined(NDEBUG)
			printf("line %d  field %d <%s>\n", iLine, iCurrFld, pcFld);
#endif
			if(iLine != iPrevLine)
			{
				if(iFldsTranslated != queue_length(q))
				{
					fprintf(stderr,
						"Line %d only contained %d (requested) fields\n",
						iPrevLine, iFldsTranslated);
					exit(1);
				}

				iFldsTranslated = 0;
				iPrevLine = iLine;

				/* rewind queue iterator */
				qitdstry(&qitTq);
				qitTq = qitcreat(q);
				ptq = qitnext(qitTq);
				if(NULL == ptq)
				{
					fprintf(stderr, "%s: (internal) Queue header lost\n", ProgName);
					exit(1);
				}

				/* write translation to COBOL file */
				if(!fLineHasError)
					ffwrite(pbRecBuf, (size_t) TrnHdr.ulRecLen, fpCobolFile);
				fLineHasError = false;
			}

			/* advance translation pointer, if necessary */
			if(NULL != ptq && ptq->iFldOffset < iCurrFld)
				ptq = qitnext(qitTq);

			/* if there's no translation for this field */
			if(NULL != ptq && ptq->iFldOffset > iCurrFld)
				continue;

			if(ptq != NULL && ptq->iFldOffset == iCurrFld)
			{
				if(!(*ptq->toCobol)(pbRecBuf, pcFld, ptq))
				{
					fprintf(stderr,
						"Translation failure on line %d field %d: '%s'\n",
						iLine, iCurrFld, pcFld);
					fLineHasError = true;
				}
				iFldsTranslated++;
			}
		}
		/* write last line, if necessary */
		if(0 < iFldsTranslated)
		{
			if(iFldsTranslated != queue_length(q))
			{
				fprintf(stderr,
					"Line %d only contained %d (requested) fields\n",
					iPrevLine, iFldsTranslated);
				exit(1);
			}
			/* write translation to COBOL file */
			if(!fLineHasError)
				ffwrite(pbRecBuf, (size_t) TrnHdr.ulRecLen, fpCobolFile);
		}
		qitdstry(&qitTq);
	}
	else
	{
		writeDmHeader(fpDmFile, q);

#if !defined(NDEBUG)
	FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
		printf("%s  fldOff %d  %ld  %ld\n", ptq->DmName, ptq->iFldOffset,
			ptq->ste.length, ptq->ste.offset);
	exit(1);
#endif
		iLine = 1;
/**************/
                bl = 4;
                counter = 0;
		while( !feof( fpCobolFile ) )
		{
		/*  counter++; */
		/* if (counter > 161)
		    break; */

		  if ( bl == 4 )
		    if ( (bl = get_len( fpCobolFile )) == EOF)
		      break;
/* printf( "bl = %d\n", bl ); */
		  if ( (rl = get_len( fpCobolFile )) == EOF )
		    break;
/* printf( "rl = %d\n", rl ); */
		  fread( pbRecBuf, 1, rl-4, fpCobolFile );
		  bl -= rl;
/**************/                     
			/* for each translation */
		  iCurrFld = 1;
		  FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
			{
				/* invoke translation function */
				if(!(*ptq->toDm)(fpDmFile, pbRecBuf, ptq))
				{
					fprintf(stderr, "Translation failure on line %d field %d\n",
						iLine, iCurrFld);
					fFldTransErr = true;
				}
				iCurrFld++;
			}
			if ( !fFldTransErr )
			{
				ffwrite( (byte *)get_dyna_buf_beg(db), get_dyna_buf_len(db),
			                                         fpDmFile );
				fputs( "\n", fpDmFile );
				(void) clear_dyna_buffer( db );
			}
			fFldTransErr = false;
         memcpy( prevIsamRecKey, pbRecBuf + TrnHdr.k.ulOffset,
                 (size_t) TrnHdr.k.ulLen ); /* update prev Isam key */
			iLine++;
		}
	}

	(void) endConversions();
	qdestroy(&q);
	free(pbRecBuf);
	free(PcIsamRec);
	free(pcFld);
	fclose(fpTrn);
	fclose(fpCobolFile);
	fclose(fpDmFile);
	return(0);
}

void usage()
{
	fprintf(stderr,
		"usage: %s [-r] [-p] <tran_filename> <cobol_filename> <dm_filename>\n",
		ProgName);
	exit(1);
}


void ffwrite(byte *pbBuf, size_t sLen, FILE *pf)
{
	if(1 != fwrite(pbBuf, sLen, 1, pf))
	{
		fprintf(stderr, "fwrite failure");
		exit(1);
	}
}


bool trnFileRead(FILE *fpTrn, char *strFname, QUEUE *pq, byte **ppbRecBuf)
{
	char line[MAX_LINE];
	char *pStart, *pEnd;
	bool parseSuccessful = true;
	QUEUE q;
	int iLineCt;
	char strCobolName[MAX_COBOL_NAME + 1];
	char strConvFuncName[20];
	tranQEntry *ptq, *ptqLast = NULL;
	byte *pbRecBuf = *ppbRecBuf;
	ulong ulLastRecOffset = 0;

	if(!fgets(line, sizeof(line), fpTrn))
	{
		fprintf(stderr, "%s empty translation request file\n", ProgName);
		return(false);
	}
	pStart = line;
	if(parseSuccessful)
		parseSuccessful = getDecimalNumber(&pStart, &TrnHdr.ulRecLen);
	if(parseSuccessful)
		parseSuccessful = getDecimalNumber(&pStart, &TrnHdr.ulMaxIsamRecLen);
	if(parseSuccessful)
		parseSuccessful = getCobolName(&pStart, TrnHdr.k.name);
	if(parseSuccessful)
		parseSuccessful = getDecimalNumber(&pStart, &TrnHdr.k.ulOffset);
	if(parseSuccessful)
		parseSuccessful = getDecimalNumber(&pStart, &TrnHdr.k.ulLen);
	if(parseSuccessful)
		parseSuccessful = getDecimalNumber(&pStart, &TrnHdr.ulFldLen);
	if(parseSuccessful)
	{
		pEnd = pStart;
		while(*pEnd && isspace(*pEnd))
			pEnd++;
		if(*pEnd)
		{
			fprintf(stderr, "%s: extraneous header value(s) in file %s: %s\n",
				ProgName, strFname, pStart);
			return(false);
		}
	}
	else
	{
		fprintf(stderr, "%s: bad transaction request file header, file %s\n",
				ProgName, strFname);
		return(false);
	}

#if !defined(NDEBUG)
   if ( pc_mode )
	   TrnHdr.keyPrefixLen = (size_t) (MAX_FILENAME_LEN + TrnHdr.k.ulLen
		                             + MAX_KEY_LEN);
   else
	   TrnHdr.keyPrefixLen = (size_t) (MAX_FILENAME_LEN + TrnHdr.k.ulLen
		                             + MAX_KEY_LEN + MAX_KEY_SUFFIX_LEN);
#else
   if ( pc_mode )
	   TrnHdr.keyPrefixLen = (size_t) (TrnHdr.k.ulLen + MAX_KEY_LEN);
   else
	   TrnHdr.keyPrefixLen = (size_t) (TrnHdr.k.ulLen
                                   + MAX_KEY_LEN + MAX_KEY_SUFFIX_LEN);
#endif

	pbRecBuf = malloc((size_t) TrnHdr.ulRecLen);
	if(NULL == pbRecBuf)
	{
		fprintf(stderr, "unable to allocate buffer of size %lu\n",
			TrnHdr.ulRecLen);
		exit(1);
	}
	memset(pbRecBuf, '\0', (size_t) TrnHdr.ulRecLen);

	q = qcreate();
	for(iLineCt = 0; fgets(line, sizeof(line), fpTrn); iLineCt++)
	{
		tranQEntry tqe;

		/* parse line, enqueue field translations, flag key field */
		pStart = line;
		/* DM name */
		parseSuccessful = getDMName(&pStart, tqe.DmName);
		/* COBOL name */
		if(parseSuccessful)
			parseSuccessful = getCobolName(&pStart, strCobolName);
		/* offset */
		if(parseSuccessful)
			parseSuccessful = getDecimalNumber(&pStart, &tqe.ste.offset);
		/* length */
		if(parseSuccessful)
			parseSuccessful = getDecimalNumber(&pStart, &tqe.ste.length);
		/* generic conv func */
		if(parseSuccessful)
			parseSuccessful = getString(&pStart, strConvFuncName,
				sizeof(strConvFuncName));
		if(parseSuccessful)
			setConvFuncPtrs(&tqe, strConvFuncName);
		/* attribute (optional) */
		if(parseSuccessful)
		{
			tqe.ste.iAttrCt = 0;
			tqe.ste.attr.type = unknownAttribute;
			if(getAttribute(&pStart, &tqe.ste.attr))
				tqe.ste.iAttrCt++;
		}
		tqe.iFldOffset = 0;

		/* check to see that line is completely parsed */
		if(parseSuccessful)
		{
			pEnd = pStart;
			while(*pEnd && isspace(*pEnd))
				pEnd++;
			if(*pEnd)
				fprintf(stderr, "%s: unrecognized attribute(s) in line %d: %s\n",
					ProgName, iLineCt, pStart);
			else
			{
				if((tqe.ste.offset + tqe.ste.length) > ulLastRecOffset)
					ulLastRecOffset = tqe.ste.offset + tqe.ste.length;
				ptq = qadd(q, sizeof(tranQEntry));
				/* track last field */
				if((tqe.ste.offset + tqe.ste.length) > ulLastRecOffset)
				{
					ulLastRecOffset = tqe.ste.offset + tqe.ste.length;
					ptqLast = ptq;
				}
				*ptq = tqe;
			}
		}
	}
	if(ulLastRecOffset > TrnHdr.ulRecLen)
	{
		fprintf(stderr,
			"The last field (%s) ends at byte %ld,"
			" but the record is only %ld bytes long\n",
			ptqLast->DmName, ulLastRecOffset, TrnHdr.ulRecLen);
		exit(1);
	}
	(void) chkFldOverlap(q, pbRecBuf);

	*pq = q;
	*ppbRecBuf = pbRecBuf;
	return(true);
}


tranQEntry *getTranQEntry(QUEUE q, char *strFldName)
{
	tranQEntry *ptq;

	FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
		if(0 == stricmp(strFldName, ptq->DmName))
			return(ptq);
	return(NULL);
}

/* test for field overlap using buffer pbRecBuf */
bool chkFldOverlap(QUEUE q, byte *pbRecBuf)
{
	QUEUE_ITERATOR qit, qitErr;
	tranQEntry *ptq, *ptq1;
	bool fErr = false;

	memset(pbRecBuf, '\0', (size_t) TrnHdr.ulRecLen);
	qit = qitcreat(q);
	for(ptq = qitnext(qit); NULL != ptq; ptq = qitnext(qit))
	{
		if(memchr(pbRecBuf + ptq->ste.offset, 1, (size_t) ptq->ste.length))
		{
				fErr = true;
				qitErr = qitcreat(q);
				for(ptq1 = qitnext(qitErr); NULL != ptq1; ptq1 = qitnext(qitErr))
				{
					if(ptq1 != ptq
					&& (
							(ptq1->ste.offset <= ptq->ste.offset
					    	&& (ptq1->ste.offset + ptq1->ste.length - 1)
						   	>= ptq->ste.offset)
						||
					   	(ptq1->ste.offset >= ptq->ste.offset
							&& ptq1->ste.offset
								<= (ptq->ste.offset + ptq->ste.length - 1))
						)
					)
						fprintf(stderr, "Field %s overlaps with field %s\n",
							ptq->DmName, ptq1->DmName);
				}
		}
		memset(pbRecBuf + ptq->ste.offset, 1, (size_t) ptq->ste.length);
	}
	memset(pbRecBuf, '\0', (size_t) TrnHdr.ulRecLen);
	return(fErr);
}


bool copyToDm(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	byte *pc = pbRecBuf + ptq->ste.offset, *pc1;
	size_t iLen = (size_t) ptq->ste.length, iLen1;
	char strDate[9];
	static char strDash[] = "-";
	attribute *pa = &ptq->ste.attr;

        ebc2asc( pc, iLen );
	fieldAdjust(&pc, &iLen);

	/* Allow only a single token.           */
        /* NOTE: If you remove the comments from the next statement
                 then you will get a name/title... with only the first
                 part of it until a blank printed, and that will be
                 it. The rest of the field will be ignored. So leave
                 the comment to get the whole thing printed out.
                 9/14/94.
        ***********************/

	for(iLen1 = 0;  /* ' ' != pc[iLen1] && */  iLen1 < iLen; iLen1++)
		;
	iLen = iLen1;

	switch(pa->type)
	{

	case dateAttribute:
   /* The following IF statement will handle Expert Input File changes
      and zero filling date fields  IF 4/25/90  PR #691
   */
   if (pc[2] == ' ')
     {
      strcpy(strDate,"00/00/00");
     }
   else
   {
		strDate[0] = pc[2];
		strDate[1] = pc[3];
		strDate[2] = '/';
		strDate[3] = pc[4];
		strDate[4] = pc[5];
		strDate[5] = '/';
		strDate[6] = pc[0];
		strDate[7] = pc[1];
		strDate[8] = EOS;
   }
		pc = strDate;
		iLen = 8;
		break;
	case tableAttribute:
		pc = getTableTranslation(pa->u.tb.tblName, pc, iLen);
		if(0 == strcmp("", pc))
			pc = pa->u.tb.strElse;
		iLen = strlen(pc);
		break;

	case validateAttribute:
		pc = getTableTranslation(pa->u.val.tblName, pc, iLen);
		if(0 == strcmp("", pc))
			pc = pa->u.val.strElse;
		iLen = strlen(pc);
		break;

	case validateIndirectOffAttribute:
		/* point to field containing table code */
		pc1 = pbRecBuf + pa->u.valindOff.ulOffset;
		iLen1 = (size_t) pa->u.valindOff.ulLen;
		fieldAdjust(&pc1, &iLen1);

		/* pc1 := name of table to be used for translating this field */
		pc1 = getTableTranslation(pa->u.valindOff.tblName, pc1, iLen1);
		if(0 == strcmp("", pc1))
			pc = pa->u.valindOff.strElse;
		else
		{
			/* pc := translation of this field */
			pc = getTableTranslation(pc1, pc, iLen);
			if(0 == strcmp("", pc))
				pc = pa->u.valindOff.strElse;
		}
		iLen = strlen(pc);
		break;
	}

	if(0 == iLen)
	{
		pc = strDash;
		iLen = strlen(pc);
	}



   (void) append_dyna_buffer( db, pc, iLen );

   (void) append_dyna_buffer( db, "|", sizeof(space) );

/********
IF space after the | causes a problem, then remove it. If it is
removed you will end up with fields separated by | only, no space.
To remove the space, delete or comment out the next statement.
3/14/94.
********/

/***   (void) append_dyna_buffer( db, &space, sizeof(space) ); ***/

	return(true);
}


bool copyToCobol(byte *pbRecBuf, char *strFld, tranQEntry *ptq)
{
	size_t sLen;

	sLen = strlen(strFld);
	if(sLen > (size_t) ptq->ste.length)
		sLen = (size_t) ptq->ste.length;
	memcpy(pbRecBuf + ptq->ste.offset, strFld, sLen);

/****** was added in March 14, 94 to act as a separator.  *******/

/******* memcpy(pbRecBuf + ptq->ste.offset, ',', 1); *********/

	/* pad with ' ' */
	if(sLen < (size_t) ptq->ste.length)
		memset(pbRecBuf + ptq->ste.offset + sLen, ' ',
			(size_t) ptq->ste.length - sLen);

	return(true);
}


int iQIFComp(char *strFileName, isamData *pID)
{
	return(stricmp(strFileName, pID->strFileName));
}

bool  startConversions()
{
   db = create_dyna_buffer( 0x1000, 0x200 );
   return(true);
}
   
bool endConversions()
{
	isamData *pID;

	if(NULL != qIsamFiles)
	{
		FOR_EACH_Q_ELE(qIsamFiles, pID, isamData *)
			fclose(pID->pF);
	}

   destroy_dyna_buffer( db );
	return(true);
}

int ksIntCompn(ulong ul0, char *s, size_t n)
{
	char format[10];
	ulong ul1;

	sprintf(format, "%%%dd", n);
	sscanf(s, format, &ul1);
	return((int) ul0 - (int) ul1);
}

/* trim leading and trailing blanks from a field starting at *ppc, extending
	for *ps characters.
 */
void fieldAdjust(char **ppc, size_t *ps)
{
	char *pc = *ppc;
	size_t s = *ps;
	/* trim leading blanks */
	while((' ' == *pc || '_' == *pc) && 0 < s)
	{
		pc++;
		s--;
	}
	*ppc = pc;	/* new starting position */

	/* trim trailing blanks */
	pc += s - 1;
	while((' ' == *pc || '_' == *pc) && 0 < s)
	{
		pc--;
		s--;
	}
	*ps = s;		/* new length */
}


bool stringToIsam(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	isamData *pID;
	byte *pc = pbRecBuf + ptq->ste.offset;
	bool bPass = true;
	attribute *pa = &ptq->ste.attr;
	size_t iLen = (size_t) ptq->ste.length;
	char *pcFileName = pa->u.ks.filename;
	char *pcKeyValue = pa->u.ks.keyvalue;

	fieldAdjust(&pc, &iLen);

	/* test if this field generates any output */
	switch(pa->type)
	{
	case keyedStringIfAttribute:
		switch(pa->u.ksif.r)
		{
		case eqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 == strncmp(pa->u.ksif.value, pc, iLen));
			break;

		case neqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 != strncmp(pa->u.ksif.value, pc, iLen));
			break;
		}
		pc = pa->u.ksif.strElse;
		iLen = strlen(pc);
		pcFileName = pa->u.ksif.filename;
		pcKeyValue = pa->u.ksif.keyvalue;
		break;

	case keyedStringIf2Attribute:
		switch(pa->u.ksif2.r)
		{
		case eqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 == strncmp(pa->u.ksif2.value, pc, iLen));
			break;

		case neqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 != strncmp(pa->u.ksif2.value, pc, iLen));
			break;
		}
		pcFileName = pa->u.ksif2.filename;
		pcKeyValue = pa->u.ksif2.keyvalue;
		break;

	case keyedStringIfIntAttribute:
		switch(pa->u.ksifInt.r)
		{
		case eqRelop:
			bPass = (bool) (0 == ksIntCompn(pa->u.ksifInt.value, pc, iLen));
			break;

		case neqRelop:
			bPass = (bool) (0 != ksIntCompn(pa->u.ksifInt.value, pc, iLen));
			break;
		}
		pc = pa->u.ksifInt.strElse;
		iLen = strlen(pc);
		pcFileName = pa->u.ksifInt.filename;
		pcKeyValue = pa->u.ksifInt.keyvalue;
		break;

	case keyedStringTableAttribute:
      /* Empty (Blank) fields are a special case, as no value should be */
      /* returned.  This could be expanded later to allow blank values to */
      /* be specified in the tables (therefore translated) */
      if(0 == iLen)
         bPass = false;
      else
      {
		   pc = getTableTranslation(pa->u.kstab.tblName, pc, iLen);
		   if(0 == strcmp("", pc))
   		   pc = pa->u.kstab.strElse;
		   iLen = strlen(pc);
		   if(0 == strcmp("", pc))
			   bPass = false;
      }
		pcFileName = pa->u.kstab.filename;
		pcKeyValue = pa->u.kstab.keyvalue;
		break;
	}

	if(!bPass)
		return(true);

	/* check count instead? */
	if(ptq->ste.attr.type)
		if(NULL == qIsamFiles)
			qIsamFiles = qcreate();

	if(NULL == (pID = qfind(qIsamFiles, pcFileName, iQIFComp)))
	{
		pID = qadd(qIsamFiles, sizeof(isamData));
		if(NULL == pID)
		{
			fprintf(stderr, "can't allocate isamData struct\n");
			exit(1);
		}
		strcpy(pID->strFileName, pcFileName);
		pID->pF = openKeyedStringFile( pID->strFileName );
		if(NULL == pID->pF)
		{
			fprintf(stderr, "can't open '%s'\n", pID->strFileName);
			exit(1);
		}
	}

	return( writeIsamRecord( pID, pbRecBuf, pa, pc, iLen ) );

} /* stringToIsam() */


bool packToIsam(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	isamData *pID;
	byte *pc = pbRecBuf + ptq->ste.offset;
	bool bPass = true;
	attribute *pa = &ptq->ste.attr;
	size_t iLen = (size_t) ptq->ste.length;
	char *pcFileName = pa->u.ks.filename;
	char *pcKeyValue = pa->u.ks.keyvalue;
	tranQEntry tqe;
	int i0;
	char strBuf[20];

	fieldAdjust(&pc, &iLen);

	tqe.ste.offset = pc - pbRecBuf;
	tqe.ste.length = iLen;
	tqe.ste.iAttrCt = 0;
	tqe.ste.attr.type = unknownAttribute;
	i0 = packToInteger(pbRecBuf, &tqe);
	sprintf(strBuf, "%d", i0);
	pc = strBuf;
	iLen = strlen(strBuf);

	/* test if this field generates any output */
	switch(pa->type)
	{
	case keyedStringIfAttribute:
		switch(pa->u.ksif.r)
		{
		case eqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 == strncmp(pa->u.ksif.value, pc, iLen));
			break;

		case neqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 != strncmp(pa->u.ksif.value, pc, iLen));
			break;
		}
		pc = pa->u.ksif.strElse;
		iLen = strlen(pc);
		pcFileName = pa->u.ksif.filename;
		pcKeyValue = pa->u.ksif.keyvalue;
		break;

	case keyedStringIf2Attribute:
		switch(pa->u.ksif2.r)
		{
		case eqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 == strncmp(pa->u.ksif2.value, pc, iLen));
			break;

		case neqRelop:
			if(0 == iLen)
				bPass = false;
			else
   			bPass = (bool) (0 != strncmp(pa->u.ksif2.value, pc, iLen));
			break;
		}
		pcFileName = pa->u.ksif2.filename;
		pcKeyValue = pa->u.ksif2.keyvalue;
		break;

	case keyedStringIfIntAttribute:
		switch(pa->u.ksifInt.r)
		{
		case eqRelop:
			bPass = (bool) (0 == ksIntCompn(pa->u.ksifInt.value, pc, iLen));
			break;

		case neqRelop:
			bPass = (bool) (0 != ksIntCompn(pa->u.ksifInt.value, pc, iLen));
			break;
		}
		pc = pa->u.ksifInt.strElse;
		iLen = strlen(pc);
		pcFileName = pa->u.ksifInt.filename;
		pcKeyValue = pa->u.ksifInt.keyvalue;
		break;

	case keyedStringTableAttribute:
		pc = getTableTranslation(pa->u.kstab.tblName, pc, iLen);
		iLen = strlen(pc);
		if(0 == strcmp("", pc))
			bPass = false;
		pcFileName = pa->u.kstab.filename;
		pcKeyValue = pa->u.kstab.keyvalue;
		break;
	}

   if(!bPass)
		return(true);

	/* check count instead? */
	if(ptq->ste.attr.type)
		if(NULL == qIsamFiles)
			qIsamFiles = qcreate();

	if(NULL == (pID = qfind(qIsamFiles, pcFileName, iQIFComp)))
	{
		pID = qadd(qIsamFiles, sizeof(isamData));
		if(NULL == pID)
		{
			fprintf(stderr, "can't allocate isamData struct\n");
			exit(1);
		}
		strcpy(pID->strFileName, pcFileName);
		pID->pF = openKeyedStringFile( pID->strFileName );
		if(NULL == pID->pF)
		{
			fprintf(stderr, "can't open '%s'\n", pID->strFileName);
			exit(1);
		}
	}

	return( writeIsamRecord( pID, pbRecBuf, pa, pc, iLen ) );

} /* packToIsam() */


FILE *openKeyedStringFile( char *fileName )
{
   if ( pc_mode )
      return ( fopen( fileName, "w" ) );
   else
      return ( fopen( fileName, "wb,type=RECORD" ) );

} /* openKeyedStringFile() */


bool  writeIsamRecord( isamData *pID, byte *pbRecBuf,
                       attribute *pa, byte *val, size_t valLen )
{
   byte *pcDest;
	char *pcFileName;
	char *pcKeyValue;
   static uint keyUniqueSuffix = 1;    /* to help form unique key suffix */

   /* needn't write empty values */
   if ( valLen == 0 )
      return ( true );

	pcFileName = pa->u.ks.filename;
	pcKeyValue = pa->u.ks.keyvalue;

   if ( pc_mode )
   {
      /* +1 for space to separate tokens */
	   memset( PcIsamRec, ' ', TrnHdr.keyPrefixLen + valLen + 1 );
	   pcDest = PcIsamRec;
#if !defined(NDEBUG)
	   /* filename */
	   memcpy( pcDest, pcFileName, strlen(pcFileName) );
	   pcDest += MAX_FILENAME_LEN;
#endif
	   /* appid */
	   memcpy( pcDest, pbRecBuf + TrnHdr.k.ulOffset, (size_t) TrnHdr.k.ulLen );
	   pcDest += TrnHdr.k.ulLen + 1; /* +1 for space-separator */
	   /* key */
	   memcpy( pcDest, pcKeyValue, strlen(pcKeyValue) );
	   pcDest += strlen(pcKeyValue) + 1;
      /* value */
	   memcpy( pcDest, val, valLen );
	   pcDest += valLen;
      *pcDest++ = (byte) '\n';
	   if( fwrite(PcIsamRec, (size_t) (pcDest - PcIsamRec), 1, pID->pF) != 1 )
		   return ( false );
      return ( true );
   }
   else
   {
	   memset( PcIsamRec, '\0', TrnHdr.keyPrefixLen + valLen );
	   pcDest = PcIsamRec;
#if !defined(NDEBUG)
	   /* filename */
	   memcpy( pcDest, pcFileName, strlen(pcFileName) );
	   pcDest += MAX_FILENAME_LEN;
#endif
	   /* appid */
	   memcpy( pcDest, pbRecBuf + TrnHdr.k.ulOffset, (size_t) TrnHdr.k.ulLen );
      if ( memcmp( prevIsamRecKey, pcDest, (size_t) TrnHdr.k.ulLen ) )
      {
         keyUniqueSuffix = 1; /* appid changed, re-start suffix from 1 */
         memcpy( prevIsamRecKey, pcDest, (size_t) TrnHdr.k.ulLen );
      }
      else
         ++keyUniqueSuffix;   /* appid unchanged, increment suffix number */
	   pcDest += TrnHdr.k.ulLen;
	   /* key */
	   memcpy( pcDest, pcKeyValue, strlen(pcKeyValue) );
	   pcDest += MAX_KEY_LEN;
	   /* add suffix to make key unique for VSAM look-up */
	   sprintf( pcDest, "%*.*u",
               MAX_KEY_SUFFIX_LEN, MAX_KEY_SUFFIX_LEN, keyUniqueSuffix );
	   pcDest += MAX_KEY_SUFFIX_LEN;
      /* value */
	   memcpy( pcDest, val, valLen );
	   pcDest += valLen;
      *pcDest = (byte) '\n';  /* despite record-orientation, need term. \n */
	   if ( fwrite(PcIsamRec, TrnHdr.keyPrefixLen + valLen + 1, 1, pID->pF)
         != 1 )
		   return ( false );
      return ( true );
   }
      
} /* writeIsamRecord() */


bool zoneToFloat(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	char strBuf[20];
	byte b;
	int iDigitCt, i, iIdx = 0;

	/* extract sign (high nibble of last byte) */
	b = ((*(pbRecBuf + ptq->ste.offset + ptq->ste.length - 1) & 0xf0) >> 4);
	switch(b)
	{
	case 0x0a:
	case 0x0c:
	case 0x0e:
	case 0x0f:
		strBuf[iIdx++] = '+';
		break;

	case 0x0b:
	case 0x0d:
		strBuf[iIdx++] = '-';
		break;
	}

	iDigitCt = (int) ptq->ste.length;
	for(i = 0; i < iDigitCt; i++)
	{
		/* insert decimal point */
		if(decimalPlacesAttribute == ptq->ste.attr.type
			&& (i == iDigitCt - (int) ptq->ste.attr.u.fd.decimalPlaces))
			strBuf[iIdx++] = '.';
				
		b = pbRecBuf[ptq->ste.offset + i] & 0x0f;
		/* both ASCII and EBCDIC have continuous digit sequences */
		strBuf[iIdx++] = b + '0';
	}
	strBuf[iIdx] = EOS;

   (void) append_dyna_buffer( db, strBuf, strlen(strBuf) );
   (void) append_dyna_buffer( db, &space, sizeof(space) );
	return(true);
}


bool floatToZone(byte *pbRecBuf, char *strFld, tranQEntry *ptq)
{
	byte *pc = strFld, *pcDest;
	byte sign;
	int i, iDigits, iLeadingZeroes;
	uint uFracDigits;

	if('-' == *pc)
	{
		sign = 0x0d;
		pc++;
	}
	else if('+' == *pc)
	{
		sign = 0x0c;
		pc++;
	}
	else
		sign = 0x0c;

	/* count int digits in the input field */
	for(iDigits = 0; isdigit(pc[iDigits]); iDigits++)
		;

	iLeadingZeroes = (int) (ptq->ste.length - iDigits);
	if(decimalPlacesAttribute == ptq->ste.attr.type)
		iLeadingZeroes -= (int) ptq->ste.attr.u.fd.decimalPlaces;

	/* check to see that integer portion fits */
	if(0 > iLeadingZeroes)
		return(false);

	/* insert leading zeroes */
	pcDest = pbRecBuf + ptq->ste.offset;
	for(i = 0; i < iLeadingZeroes; i++)
		*pcDest++ = 0xf0;

	/* copy integer portion */
	while(isdigit(*pc))
		*pcDest++ = (byte) 0xf0 | (*pc++ - '0');

	/* skip decimal point */
	if('.' == *pc)
		pc++;

	if(decimalPlacesAttribute == ptq->ste.attr.type)
		uFracDigits = (uint) ptq->ste.attr.u.fd.decimalPlaces;
	else
		uFracDigits = 0;
	/* copy fractional portion; excess truncated */
	for(i = 0; i < uFracDigits; i++)
		if(*pc)
			*pcDest++ = (byte) (0xf0 | (*pc++ - '0'));
		else
			*pcDest++ = 0xf0;

	/* insert sign */
	pcDest--;
	*pcDest = (*pcDest & 0x0f) | (sign << 4);

	return(true);
}

/* fractions are truncated */
int zoneToInteger(byte *pbRecBuf, tranQEntry *ptq)
{
	char strBuf[20];
	byte b;
	int iDigitCt, i, iIdx = 0;
	bool bDecimalEncountered = false;
	attribute *pat = &ptq->ste.attr;

	/* extract sign (high nibble of last byte) */
	b = ((*(pbRecBuf + ptq->ste.offset + ptq->ste.length - 1)) & 0xf0) >> 4;
	switch(b)
	{
	case 0x0a:
	case 0x0c:
	case 0x0e:
	case 0x0f:
		strBuf[iIdx++] = '+';
		break;

	case 0x0b:
	case 0x0d:
		strBuf[iIdx++] = '-';
		break;
	}

	iDigitCt = (int) ptq->ste.length;
	for(i = 0; i < iDigitCt && !bDecimalEncountered; i++)
	{
		/* stop at decimal point */
		if(decimalPlacesAttribute == pat->type
			&& (i == iDigitCt - (int) pat->u.fd.decimalPlaces))
			bDecimalEncountered = true;
		if(!bDecimalEncountered)
		{
			b = pbRecBuf[ptq->ste.offset + i] & 0x0f;
			/* both ASCII and EBCDIC have continuous digit sequences */
			strBuf[iIdx++] = b + '0';
		}
	}
	strBuf[iIdx] = EOS;
	return(atoi(strBuf));
}

/* fractions are truncated */
bool zoneToInt(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	char strBuf[20];
	byte b, *pb;
	int iDigitCt, i, iIdx = 0, i0, i1;
	bool bDecimalEncountered = false;
	attribute *pat = &ptq->ste.attr;
	tranQEntry tqe;
	long l0;

	/* extract sign (high nibble of last byte) */
	b = ((*(pbRecBuf + ptq->ste.offset + ptq->ste.length - 1)) & 0xf0) >> 4;
	switch(b)
	{
	case 0x0a:
	case 0x0c:
	case 0x0e:
	case 0x0f:
		strBuf[iIdx++] = '+';
		break;

	case 0x0b:
	case 0x0d:
		strBuf[iIdx++] = '-';
		break;
	}

	iDigitCt = (int) ptq->ste.length;
	for(i = 0; i < iDigitCt && !bDecimalEncountered; i++)
	{
		/* stop at decimal point */
		if(decimalPlacesAttribute == pat->type
			&& (i == iDigitCt - (int) pat->u.fd.decimalPlaces))
			bDecimalEncountered = true;
		if(!bDecimalEncountered)
		{
			b = pbRecBuf[ptq->ste.offset + i] & 0x0f;
			/* both ASCII and EBCDIC have continuous digit sequences */
			strBuf[iIdx++] = b + '0';
		}
	}
	strBuf[iIdx] = EOS;

	switch(pat->type)
	{
	case by12OffAttribute:
		tqe.ste.offset = pat->u.by12Off.ulOff;
		tqe.ste.length = pat->u.by12Off.ulLen;
		tqe.ste.iAttrCt = 0;
		tqe.ste.attr.type = unknownAttribute;
		i0 = zoneToInteger(pbRecBuf, &tqe);
		i1 = atoi(strBuf);
		sprintf(strBuf, "%d", i0 * 12 + i1);
		break;

	case byThousandAttribute:
		l0 = atol(strBuf);
		sprintf(strBuf, "%ld", l0 / 1000);
		break;

	case toMonthsAttribute:
		pb = strBuf;
		if('+' == *pb || '-' == *pb)
			pb++;
		i0 = 10 * (pb[0] - '0') + (pb[1] - '0');
		i1 = 10 * (pb[2] - '0') + (pb[3] - '0');
		sprintf(strBuf, "%d", 12 * i0 + i1);
		break;
	}

   (void) append_dyna_buffer( db, strBuf, strlen(strBuf) );
   (void) append_dyna_buffer( db, &space, sizeof(space) );
	return(true);
}


bool intToZone(byte *pbRecBuf, char *strFld, tranQEntry *ptq)
{
	byte *pc = strFld, *pcDest;
	byte sign;
	int i;
	unsigned int uDigits;
	int iLeadingZeroes;
	uint uFracDigits;

	if('-' == *pc)
	{
		sign = 0x0d;
		pc++;
	}
	else if('+' == *pc)
	{
		sign = 0x0c;
		pc++;
	}
	else
		sign = 0x0c;

	/* count int digits in the input field */
	for(uDigits = 0; isdigit(pc[uDigits]); uDigits++)
		;

	iLeadingZeroes = (int) (ptq->ste.length - uDigits);
	if(decimalPlacesAttribute == ptq->ste.attr.type)
		iLeadingZeroes -= (int) ptq->ste.attr.u.fd.decimalPlaces;

	/* check to see that integer portion fits */
	if(0 > iLeadingZeroes)
		return(false);

	/* insert leading zeroes */
	pcDest = pbRecBuf + ptq->ste.offset;
	for(i = 0; i < iLeadingZeroes; i++)
		*pcDest++ = 0xf0;

	/* copy integer portion */
	while(isdigit(*pc))
		*pcDest++ = (byte) (0xf0 | (*pc++ - '0'));

	if(decimalPlacesAttribute == ptq->ste.attr.type)
		uFracDigits = (uint) ptq->ste.attr.u.fd.decimalPlaces;
	else
		uFracDigits = 0;
	/* insert trailing zeroes */
	for(i = 0; i < uFracDigits; i++)
		*pcDest++ = 0xf0;

	/* insert sign */
	pcDest--;
	*pcDest = (*pcDest & 0x0f) | (sign << 4);

	return(true);
}


bool packToFloat(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	char strBuf[20];
	byte b;
	int iDigitCt, i, iIdx = 0;

	/* extract sign */
	b = *(pbRecBuf + ptq->ste.offset + ptq->ste.length - 1) & 0x0f;
	switch(b)
	{
	case 0x0a:
	case 0x0c:
	case 0x0e:
	case 0x0f:
		strBuf[iIdx++] = '+';
		break;

	case 0x0b:
	case 0x0d:
		strBuf[iIdx++] = '-';
		break;
	}

	iDigitCt = (int) ptq->ste.length * 2 - 1;
	for(i = 0; i < iDigitCt; i++)
	{
		/* insert decimal point */
		if(decimalPlacesAttribute == ptq->ste.attr.type
			&& (i == iDigitCt - (int) ptq->ste.attr.u.fd.decimalPlaces))
			strBuf[iIdx++] = '.';
				
		b = *((byte *) pbRecBuf + ptq->ste.offset + i / 2);
		if(even(i))
			b = (b & 0xf0) >> 4;
		else
			b &= 0x0f;
		/* both ASCII and EBCDIC have continuous digit sequences */
		strBuf[iIdx++] = b + '0';
	}
	strBuf[iIdx] = EOS;

   (void) append_dyna_buffer( db, strBuf, strlen(strBuf) );
   (void) append_dyna_buffer( db, "|", sizeof(space) );

/********
IF space after the | causes a problem, then remove it. If it is
removed you will end up with fields separated by | only, no space.
To remove the space, delete or comment out the next statement.
3/14/94.
********/

   (void) append_dyna_buffer( db, &space, sizeof(space) );

	return(true);
}


/* Note: when writing to even nibble (high order), odd nibble
	can be overwritten.
 */
bool floatToPack(byte *pbRecBuf, char *strFld, tranQEntry *ptq)
{
	byte *pc = strFld, *pcDest;
	byte sign;
	int i, iDigits, iLeadingZeroes, iPackDigits;
	uint uFracDigits;

	if('-' == *pc)
	{
		sign = 0x0d;
		pc++;
	}
	else if('+' == *pc)
	{
		sign = 0x0c;
		pc++;
	}
	else
		sign = 0x0c;

	iPackDigits = (int) ptq->ste.length * 2 - 1;

	/* count int digits in the input field */
	for(iDigits = 0; isdigit(pc[iDigits]); iDigits++)
		;

	iLeadingZeroes = (int) (iPackDigits - iDigits);
	if(decimalPlacesAttribute == ptq->ste.attr.type)
		iLeadingZeroes -= (int) ptq->ste.attr.u.fd.decimalPlaces;

	/* check to see that integer portion fits */
	if(0 > iLeadingZeroes)
		return(false);

	/* insert leading zeroes */
	pcDest = pbRecBuf + ptq->ste.offset;
	for(i = 0; i < iLeadingZeroes; i++)
	{
		if(even(i))
			*pcDest = 0;
		else
			*pcDest++ = 0;
	}

	/* copy integer portion */
	for(i = 0; i < iDigits; i++)
	{
		if(even(i + iLeadingZeroes))
			*pcDest = (*pc++ - '0') << 4;
		else
		{
			*pcDest = (*pcDest & 0xf0) | *pc++ - '0';
			pcDest++;
		}
	}

	/* skip decimal point */
	if('.' == *pc)
		pc++;

	if(decimalPlacesAttribute == ptq->ste.attr.type)
		uFracDigits = (uint) ptq->ste.attr.u.fd.decimalPlaces;
	else
		uFracDigits = 0;
	/* copy fractional portion; excess truncated */
	for(i = 0; i < uFracDigits; i++)
	{
		byte bDigitVal;

		if(*pc)
			bDigitVal = *pc++ - '0';
		else
			bDigitVal = 0;

		if(even(iLeadingZeroes + iDigits + i))
			*pcDest = bDigitVal << 4;
		else
		{
			*pcDest = (*pcDest & 0xf0) | bDigitVal;
			pcDest++;
		}
	}

	/* insert sign */
	pcDest = pbRecBuf + ptq->ste.offset + ptq->ste.length - 1;
	*pcDest = (*pcDest & 0xf0) | sign;

	return(true);
}


/* fractions are truncated */
int packToInteger(byte *pbRecBuf, tranQEntry *ptq)
{
	char strBuf[20];
	byte b;
	int iDigitCt, i, iIdx = 0;
	bool bDecimalEncountered = false;

	/* extract sign */
	b = *(pbRecBuf + ptq->ste.offset + ptq->ste.length - 1) & 0x0f;
	switch(b)
	{
	case 0x0a:
	case 0x0c:
	case 0x0e:
	case 0x0f:
		strBuf[iIdx++] = '+';
		break;

	case 0x0b:
	case 0x0d:
		strBuf[iIdx++] = '-';
		break;
	}

	iDigitCt = (int) ptq->ste.length * 2 - 1;
	for(i = 0; i < iDigitCt && !bDecimalEncountered; i++)
	{
		/* insert decimal point */
		if(decimalPlacesAttribute == ptq->ste.attr.type
			&& (i == iDigitCt - (int) ptq->ste.attr.u.fd.decimalPlaces))
			bDecimalEncountered = true;

		if(!bDecimalEncountered)
		{
			b = *(pbRecBuf + ptq->ste.offset + i / 2);
			if(even(i))
				b = (b & 0xf0) >> 4;
			else
				b &= 0x0f;
			/* both ASCII and EBCDIC have continuous digit sequences */
			strBuf[iIdx++] = b + '0';
		}
	}
	strBuf[iIdx] = EOS;

	return(atoi(strBuf));
} /* packToInteger() */


/* fractions are truncated */
bool packToInt(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	char strBuf[20];
	byte b;
	int iDigitCt, i, iIdx = 0;
	bool bDecimalEncountered = false;
	long l;

	/* extract sign */
	b = *(pbRecBuf + ptq->ste.offset + ptq->ste.length - 1) & 0x0f;
	switch(b)
	{
	case 0x0a:
	case 0x0c:
	case 0x0e:
	case 0x0f:
		strBuf[iIdx++] = '+';
		break;

	case 0x0b:
	case 0x0d:
		strBuf[iIdx++] = '-';
		break;
	}

	iDigitCt = (int) ptq->ste.length * 2 - 1;
	for(i = 0; i < iDigitCt && !bDecimalEncountered; i++)
	{
		/* insert decimal point */
		if(decimalPlacesAttribute == ptq->ste.attr.type
			&& (i == iDigitCt - (int) ptq->ste.attr.u.fd.decimalPlaces))
			bDecimalEncountered = true;

		if(!bDecimalEncountered)
		{
			b = *(pbRecBuf + ptq->ste.offset + i / 2);
			if(even(i))
				b = (b & 0xf0) >> 4;
			else
				b &= 0x0f;
			/* both ASCII and EBCDIC have continuous digit sequences */
			strBuf[iIdx++] = b + '0';
		}
	}
	strBuf[iIdx] = EOS;

	switch(ptq->ste.attr.type)
	{
	case byThousandAttribute:
		l = atol(strBuf);
		sprintf(strBuf, "%ld", l / 1000);
		break;
	}

   (void) append_dyna_buffer( db, strBuf, strlen(strBuf) );
   (void) append_dyna_buffer( db, &space, sizeof(space) );
	return(true);
}


bool intToPack(byte *pbRecBuf, char *strFld, tranQEntry *ptq)
{
	byte *pc = strFld, *pcDest;
	byte sign;
	int i, iDigits, iLeadingZeroes, iPackDigits;
	uint uFracDigits;

	if('-' == *pc)
	{
		sign = 0x0d;
		pc++;
	}
	else if('+' == *pc)
	{
		sign = 0x0c;
		pc++;
	}
	else
		sign = 0x0c;

	iPackDigits = (int) ptq->ste.length * 2 - 1;

	/* count int digits in the input field */
	for(iDigits = 0; isdigit(pc[iDigits]); iDigits++)
		;

	iLeadingZeroes = (int) (iPackDigits - iDigits);
	if(decimalPlacesAttribute == ptq->ste.attr.type)
		iLeadingZeroes -= (int) ptq->ste.attr.u.fd.decimalPlaces;

	/* check to see that integer portion fits */
	if(0 > iLeadingZeroes)
		return(false);

	/* insert leading zeroes */
	pcDest = pbRecBuf + ptq->ste.offset;
	for(i = 0; i < iLeadingZeroes; i++)
	{
		if(even(i))
			*pcDest = 0;
		else
			*pcDest++ = 0;
	}

	/* copy integer portion */
	for(i = 0; i < iDigits; i++)
	{
		if(even(i + iLeadingZeroes))
			*pcDest = (*pc++ - '0') << 4;
		else
		{
			*pcDest = (*pcDest & 0xf0) | *pc++ - '0';
			pcDest++;
		}
	}

	/* skip decimal point */
	if('.' == *pc)
		pc++;

	if(decimalPlacesAttribute == ptq->ste.attr.type)
		uFracDigits = (uint) ptq->ste.attr.u.fd.decimalPlaces;
	else
		uFracDigits = 0;
	/* zero fractional portion */
	for(i = 0; i < uFracDigits; i++)
	{
		if(even(iLeadingZeroes + iDigits + i))
			*pcDest = 0;
		else
		{
			*pcDest &= 0xf0;
			pcDest++;
		}
	}

	/* insert sign */
	pcDest = pbRecBuf + ptq->ste.offset + ptq->ste.length - 1;
	*pcDest = (*pcDest & 0xf0) | sign;

	return(true);
}


bool unsupported(FILE *fpDmFile, byte *pbRecBuf, tranQEntry *ptq)
{
	fprintf(stderr, "unsupported operation requested on field %s\n",
		ptq->DmName);
	return(false);
}


bool unsupportedToCobol(byte *pbRecBuf, char *strFld, tranQEntry *ptq)
{
	fprintf(stderr, "unsupported operation requested on field %s\n",
		ptq->DmName);
	return(false);
}


void writeDmHeader(FILE *f, QUEUE q)
{
	tranQEntry *ptq;

	FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
	{
		/* skip strings (they'll be emitted to separate file(s)) */
		if(stringToIsam == ptq->toDm || packToIsam == ptq->toDm)
			continue;
		fputs(ptq->DmName, f);
		fputc(' ', f);
	}
	fputs("\n", f);
}

bool chkFldsAvail(QUEUE q)
{
	tranQEntry *ptq;
	int iErrCnt = 0;

	FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
	{
		if(0 == ptq->iFldOffset)
		{
			iErrCnt++;
			fprintf(stderr,
				"%s: The field '%s' is not available in the DM output\n",
				ProgName, ptq->DmName);
		}
	}
	return(0 == iErrCnt);
}


void setFldCnt(QUEUE q, char dmName[], int iFldCt)
{
	tranQEntry *ptq;

	FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
	{
		if(0 == stricmp(ptq->DmName, dmName))
			ptq->iFldOffset = iFldCt;
	}
}


void readDmHeader(FILE *f, QUEUE q)
{
	int iChar;
	char dmName[MAX_DM_NAME];
	int iIdx = 0;
	bool bInFld = false;
	int iFldCt = 0;

	strcpy(dmName, "");
	while('\n' != (iChar = fgetc(f)) && EOF != iChar)
	{
		if(isspace(iChar))
		{
			dmName[iIdx] = EOS;
			if(bInFld)
			{
				setFldCnt(q, dmName, iFldCt);
				bInFld = false;
				iIdx = 0;
			}
		}
		else
		{
			if(!bInFld)
			{
				bInFld = true;
				iFldCt++;
			}
			dmName[iIdx++] = (char) iChar;
		}
	}
	/* finish last name, if any */
	if(bInFld)
	{
		dmName[iIdx] = EOS;
		setFldCnt(q, dmName, iFldCt);
	}
}


void setConvFuncPtrs(tranQEntry *ptq, char strConvFuncName[])
{
	int i;
	bool found = false;

	for(i = 0; !found && 0 != strcmp("", convFuncTable[i].name); i++)
	{
		if(0 == stricmp(convFuncTable[i].name, strConvFuncName))
			found = true;
	}
	if(found)
	{
		i--;
		ptq->toDm = convFuncTable[i].toDm;
		ptq->toCobol = convFuncTable[i].toCobol;
	}
	else
	{
		ptq->toDm = unsupported;
		ptq->toCobol = unsupportedToCobol;
	}
}


int compTQEbyFldOrder(tranQEntry *ptq0, tranQEntry *ptq1)
{
	return(ptq0->iFldOffset - ptq1->iFldOffset);
}


QUEUE sortQueueIntoFieldOrder(QUEUE q)
{
	size_t i, iElemCt = queue_length(q);
	tranQEntry *ptq, *ptqArray, *ptqIter;
	QUEUE qNew;

	ptqArray = ptqIter = malloc(iElemCt * sizeof(tranQEntry));
	if(NULL == ptqArray)
	{
		fprintf(stderr, "Insufficient memory to sort tranQ\n");
		exit(1);
	}

	/* copy queue to array */
	FOR_EACH_Q_ELE(q, ptq, tranQEntry *)
		*ptqIter++ = *ptq;

	/* sort array */
	qsort(ptqArray, iElemCt, sizeof(tranQEntry), compTQEbyFldOrder);

	/* write array to new queue */
	qNew = qcreate();
	for(i = 0; i < iElemCt; i++)
	{
		ptq = qadd(qNew, sizeof(tranQEntry));
		*ptq = ptqArray[i];
	}
	return(qNew);
}


int getFld(FILE *fpDmFile, char *strFld, int bufSize, int *piLine,
	int *piCurrFld)
{
	int c, c1;
	static bool bNewLine = false;
	int iIdx = 0;
	bool fReadAhead = false;

	/* skip blanks */
	c = fgetc(fpDmFile);
	while(EOF != c && isspace(c))
	{
		if('\n' == c)
			bNewLine = true;
		c = fgetc(fpDmFile);
	}

	if(bNewLine)
	{
		(*piLine)++;
		*piCurrFld = 0;
		bNewLine = false;
	}

	if('"' == c)
	{
		(*piCurrFld)++;
		c = fgetc(fpDmFile);
		while(EOF != c && '"' != c && '\n' != c)
		{
			if('\\' == (char) c)
			{
				c1 = fgetc(fpDmFile);
				if('"' == (char) c1)
					c = c1;
				else
					fReadAhead = true;
			}
			/* truncate excess field length */
			if((iIdx + 2) < bufSize)
				strFld[iIdx++] = (char) c;
			if(fReadAhead)
			{
				c = c1;
				fReadAhead = false;
			}
			else
				c = fgetc(fpDmFile);
		}
		strFld[iIdx] = EOS;
		if(EOF == c || '\n' == c)
		{
			fprintf(stderr, "Unterminated string on Dm line %d\n", *piLine);
			exit(1);
		}
	}
	else
	{
		(*piCurrFld)++;
		while(EOF != c && !isspace(c))
		{
			/* truncate excess field length */
			if((iIdx + 2) < bufSize)
				strFld[iIdx++] = (char) c;
			c = fgetc(fpDmFile);
		}
		strFld[iIdx] = EOS;
		if('\n' == c)
			bNewLine = true;
	}
	return(c);
}

/**********/
int get_len( FILE *fp )
{
   int i;

   i = (unsigned) fgetc( fp );
   i = 256*i + (unsigned) fgetc( fp );

   (void) fgetc( fp );
   (void) fgetc( fp );

   if ( feof( fp ) )
      return EOF;
   else
      return i;
}

char *ebc2asc( char *pc, size_t iLen )
{
   int i;

   for ( i = 0;  i < iLen;  ++i )
      pc[i] = ebc2asc_tbl[ (unsigned char) pc[i] ];

   return pc;
}

/**********/
